import numpy as np

np.seterr(all="raise")

from algorithm.parameters import params
from utilities.fitness.get_data import get_data
import utilities.fitness.math_functions as mf
import utilities.fitness.traffic_functions as tf
from utilities.fitness.optimize_constants import optimize_constants

from fitness.base_ff_classes.base_ff import base_ff


def getgrammarconstants():
    """
    We set the extra parameters that are needed in the grammar.
    np: Numpy
    mf: utilities.fitness.math_functions
    tf: utilities.fitness.traffic_functions
    nr_phase: the number of phases in the test data
    sat_flow_rate: the identified saturated flow rate for the network
    mean_green: the mean green time of the intersections
    time: the cycle time of the network
    nr_inter: the number of intersections in the data set
    phase_time: empty variable that will be filled in the grammar
    """
    return {'np': np, 'mf': mf, 'tf': tf, 'nr_phase': 4, 'sat_flow_rate': 0.83, 'mean_green': 12, 'time': 60,
            'nr_inter': 3, 'phase_time': 0}


class four_phase(base_ff):
    """
    Fitness function for supervised learning, ie regression and
    classification problems. Given a set of training or test data,
    returns the error between y (true labels) and yhat (estimated
    labels).

    We can pass in the error metric and the dataset via the params
    dictionary. Of error metrics, eg RMSE is suitable for regression,
    while F1-score, hinge-loss and others are suitable for
    classification.

    This is an abstract class which exists just to be subclassed:
    should not be instantiated.
    """

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        # Get training and test data
        self.training_in, self.training_exp, self.test_in, self.test_exp = \
            get_data(params['DATASET_TRAIN'], params['DATASET_TEST'])

        # Find number of variables.
        self.n_vars = np.shape(self.training_in)[0]

        # Regression/classification-style problems use training and test data.
        if params['DATASET_TEST']:
            self.training_test = True

    def evaluate(self, ind, **kwargs):
        """
        Note that math functions used in the solutions are imported from either
        utilities.fitness.math_functions or called from numpy.

        :param ind: An individual to be evaluated.
        :param kwargs: An optional parameter for problems with training/test
        data. Specifies the distribution (i.e. training or test) upon which
        evaluation is to be performed.
        :return: The fitness of the evaluated individual.
        """

        dist = kwargs.get('dist', 'training')

        if dist == "training":
            # Set training datasets.
            x = self.training_in
            y = self.training_exp

        elif dist == "test":
            # Set test datasets.
            x = self.test_in
            y = self.test_exp

        else:
            raise ValueError("Unknown dist: " + dist)

        grammarConstants = getgrammarconstants()

        extra_var = {'np': np, 'mf': mf, 'tf': tf, 'nr_phase': 4, 'sat_flow_rate': 0.83, 'mean_green': 12, 'time': 60,
                     'nr_inter': 3, 'phase_time': 0}

        if params['OPTIMIZE_CONSTANTS']:
            # if we are training, then optimize the constants by
            # gradient descent and save the resulting phenotype
            # string as ind.phenotype_with_c0123 (eg x[0] +
            # c[0] * x[1]**c[1]) and values for constants as
            # ind.opt_consts (eg (0.5, 0.7). Later, when testing,
            # use the saved string and constants to evaluate.
            print("\n" + ind.phenotype)
            # g_tot = 52
            # nr_phase=4
            # loss = params['ERROR_METRIC']


            if dist == "training":
                return optimize_constants(x, y, ind)

            else:
                return optimize_constants(x, y, ind)
                # # this string has been created during training
                # phen = ind.phenotype_consec_consts
                # f = eval("lambda x: " + phen)
                # c = ind.opt_consts
                # # phen will refer to x (ie test_in), and possibly to c
                # y_hat = np.zeros([int(np.size(y)), 1])
                # for i in range(int(np.size(y) / nr_phase)):
                #     x_ind = x[:, i * nr_phase]
                #     score_q = np.zeros([nr_phase, 1])
                #     score_a = np.zeros([nr_phase, 1])
                #     for j in range(nr_phase):
                #         score_q[j] = f(x_ind[j])
                #         score_a[j] = f(x_ind[j + nr_phase])
                #     score_a_ave = np.average(score_a)
                #     score_q_ave = np.average(score_q)
                #     for j in range(nr_phase):
                #         y_hat[i * nr_phase + j] = 13 + c[0] * (score_q[j] - score_q_ave) + c[1] * (
                #                     score_a[j] - score_a_ave)
                # assert np.isrealobj(y_hat)
                #
                # # let's always call the error function with the
                # # true values first, the estimate second
                # return params['ERROR_METRIC'](y, y_hat)

        else:
            print("\n" + ind.phenotype)
            # phenotype won't refer to C

            # total phase time
            g_tot = 52

            # store parameters for easy use
            loss = params['ERROR_METRIC']
            nr_phase = extra_var['nr_phase']

            # initialize value function with zeros
            function_value_total = np.zeros([int(np.size(y) / nr_phase), 1])

            # for all the data points. One data point is one intersection
            for i in range(int(np.size(y) / nr_phase)):
                # initialize function value of the individual to an unrealistic high value
                fval_individual = 10000

                # get the values to compare with theta_1 and theta_2 in the
                # grammar. The variable x is ordered as: [q_rel', q_max', q']'.
                queue = x[2 * nr_phase:3 * nr_phase, i * nr_phase + 1]
                queue_max = x[nr_phase:2 * nr_phase, i * nr_phase + 1]

                # also store all extra variables to use in the grammar
                extra_var['x'] = x[:, i * nr_phase:(i + 1) * nr_phase]

                # try all combinations of parameters
                # for all the first parameters
                for theta_1 in queue:
                    # for all the second parameters
                    for theta_2 in queue_max:
                        # store the to-be optimized parameters in c as [theta_1', theta_2']'
                        extra_var['c'] = np.array([theta_1, theta_2])

                        # evaluate program
                        exec(ind.phenotype, extra_var)

                        # loss between y and output program
                        # alpha * (g - y)^2 in eq. 4-6 of thesis
                        fval = 0.05 * loss(y[i * nr_phase:(i + 1) * nr_phase], extra_var['phase_time'])

                        # Also add the fval after multiplying with the total green time
                        times = extra_var['phase_time']
                        if np.amin(times) < 2:
                            fval = 1000
                        else:
                            # calculate relative output
                            # equation 4-5 in thesis
                            times = np.divide(times, np.sum(times)) * g_tot

                            # loss between y and relative output
                            # other part of eq. 4-6 in thesis
                            fval += loss(y[i * nr_phase:(i + 1) * nr_phase], times)

                        # set function value of individual if this combination of
                        # theta_1/theta_2 is lower than current fval
                        if fval < fval_individual:
                            fval_individual = fval

                # store all function values for this grammar
                function_value_total[i] = fval_individual

            assert np.isrealobj(np.mean(function_value_total))

            # let's always call the error function with the true
            # values first, the estimate second
            return np.mean(function_value_total)
